#include <math.h>
#include <iostream>
#include <assert.h>
#include <new>
#define CUSTOMNEW
#include "2heap.h"
#include "list.h"

struct GraphArc {
  long source;
  long dest;
  float weight;
  void *data;
};

struct GraphState {
  List<GraphArc> arcs;
};

struct Graph {
  GraphState *states;
  long nStates;
};

Graph reverseGraph(Graph g) 
{
  Graph rev;
  rev.states = new GraphState[g.nStates];
  rev.nStates = g.nStates;
  for ( long i  = 0 ; i < g.nStates ; ++i )
    for ( ListIter<GraphArc> l(g.states[i].arcs) ; l ; ++l ) {
      GraphArc r;
      r.data = &l.data();
      r.dest = i;
      r.source = l.data().dest;
      r.weight = l.data().weight;
      rev.states[r.source].arcs.push(r);
    }
  return rev;
}

// Depth First Search (only one search can be active at once)

Graph dfsGraph;
bool *dfsVis;
void (*dfsFunc)(long, long);
void (*dfsExitFunc)(long, long);

void dfsRec(long state, long pred) {
  if ( dfsVis[state] )
    return;
  dfsVis[state] = true;
  if ( dfsFunc )
    dfsFunc(state, pred);
  for ( ListIter<GraphArc> l(dfsGraph.states[state].arcs) ; l ; ++l ) {
    long dest = l.data().dest;
    dfsRec(dest, state);
  }
  if ( dfsExitFunc )
    dfsExitFunc(state, pred);
}

inline void depthFirstSearch(Graph graph, long startState, bool* visited, void (*func)(long state, long pred)) {
  dfsGraph = graph;
  dfsVis = visited;
  dfsFunc = func;
  dfsExitFunc = NULL;
  dfsRec(startState, -1);
}

List<long> *topSort;

void pushTopo(long state, long pred) {
  topSort->push(state);
  pred = pred;			// dummy statement to avoid warnings
}

List<long> *topologicalSort(Graph g) {
  topSort = new List<long>;
  dfsGraph = g;
  dfsVis = new bool[g.nStates];
  dfsFunc = NULL;
  dfsExitFunc = pushTopo;
  long i;
  for ( i = 0 ; i < g.nStates ; ++i )
    dfsVis[i] = 0;
  for ( i = 0 ; i < g.nStates ; ++i )
    dfsRec(i, -1);
  delete[] dfsVis;
  return topSort;
}

// Dijikstra's single source shortest path tree algorithm

struct DistToState {
  // when used in a dumb packed binary heap, this structure
  // keeps track of where each state's distance is in the heap
  long state;
  static DistToState **stateLocations;
  static float *weights;
  static float unreachable;
  operator float() const { return weights[state]; }
  void operator = (DistToState rhs) { 
    stateLocations[rhs.state] = this;
    state = rhs.state;
  }
};

float *DistToState::weights = NULL;
DistToState **DistToState::stateLocations = NULL;
float DistToState::unreachable = HUGE_VAL;

inline bool operator < (DistToState lhs, DistToState rhs) {
  return DistToState::weights[lhs.state] > DistToState::weights[rhs.state];
}

inline bool operator == (DistToState lhs, DistToState rhs) {
  return DistToState::weights[lhs.state] == DistToState::weights[rhs.state];
}

inline bool operator == (DistToState lhs, float rhs) {
  return DistToState::weights[lhs.state] == rhs;
}

// fills dist[state] with the distance from state to dest
// returns a graph containing only the edges along the shortest
// paths tree.  the GraphArc.data field in the return tree
// is a pointer to the GraphArc in the original graph
Graph shortestPathTree(Graph g, long dest, float *dist)
{
  long nStates = g.nStates;
  GraphArc **best = new GraphArc *[nStates];
  
  GraphState *rev = reverseGraph(g).states;

  GraphState *pathTree = new GraphState[nStates];
  long nUnknown = nStates;

  DistToState *distQueue = new DistToState[nStates];

  float *weights = dist;
  long i;
  for ( i = 0 ; i < nStates ; ++i ) {
    weights[i] = HUGE_VAL;
  }
  
  DistToState **stateLocations = new DistToState *[nStates];
  DistToState::weights = weights;
  DistToState::stateLocations = stateLocations;
  
  weights[dest] = 0;
  for ( i = 1; i < nStates ; ++i ) {
    long fillWith;
    if ( i <= dest )
      fillWith = i-1;
    else
      fillWith = i;
    distQueue[i].state = fillWith;
    stateLocations[fillWith] = &distQueue[i];
  }
  distQueue[0].state = dest;
  stateLocations[dest] = &distQueue[0];
  
  for ( i = 0 ; i < nStates ; ++i )
    best[i] = NULL;

  float candidate;
  for ( ; ; ) {
    if ( (float)distQueue[0] == HUGE_VAL || nUnknown == 0 ) {
      break;
    }
    long targetState, activeState = distQueue[0].state;
    heapPop(distQueue, distQueue + nUnknown--);
    for ( ListIter<GraphArc> a = rev[activeState].arcs ; a ; ++a ) {
      // future: compare only best arc to any given state
      targetState = a.data().dest;
      if ( (candidate = a.data().weight + weights[activeState] )
	   < weights[targetState] ) {
	
	weights[targetState] = candidate;
	best[targetState] = (GraphArc *)a.data().data;
	heapAdjustUp(distQueue, stateLocations[targetState]);
      }
    }
  }

  for ( i = 0 ; i < nStates ; ++i )
    if ( best[i] ) {
      pathTree[i].arcs.push(*best[i]);
      pathTree[i].arcs.top().data = best[i];
    }
  
  delete[] stateLocations;
  delete[] distQueue;
  delete[] rev;
  delete[] best;

  Graph ret;
  ret.nStates = nStates;
  ret.states = pathTree;
  return ret;
}

// rudimentary graph I/O (no error checking)

ostream & operator << (ostream &o, GraphArc &a)
{
  //////return o << '(' << a.source << ' ' << a.dest << ' ' << a.weight << ')';
return o << a.source << ' ' << a.dest << ' ' << a.weight << endl;
}

istream & operator >> (istream &istr, GraphArc &a)
{
 // char c;
  long i;
 // istr >> c;			// open paren
  istr >> a.source;
  istr >> a.dest;
  istr >> a.weight;
 ////// istr >> c;			// close paren
  a.data = NULL;
  return istr;
}

istream & operator >> (istream &istr, GraphState &s)
{
  char c;
  return istr;
}

istream & operator >> (istream &istr, Graph &g)
{
 // char c;
  GraphArc a,b;
 // istr >> g.nStates;
  if ( istr && g.nStates > 0 )
    g.states = new GraphState[g.nStates];
  else
    g.states = NULL;
  for ( ; ; ) {
  ///////  istr >> c;
  //////////  if ( !istr || c != '(')
if ( !istr )
      break;
  //////  istr.putback(c);
    istr >> a;
    //a.weight=1-a.weight;//////////////////////////////////////
    b.source=a.dest;
    b.weight=a.weight;
    b.dest=a.source;
    if ( !(istr && a.source >= 0 && a.source < g.nStates) )
      break;
    g.states[a.source].arcs.push(a);
    g.states[b.source].arcs.push(b);
  }
  return istr;
}

ostream & operator << (ostream &out, Graph g)
{
 // out << g.nStates << '\n';;
  for ( long i = 0 ; i < g.nStates ; ++i ) {
    for ( ListIter<GraphArc> a(g.states[i].arcs) ; a ; ++a )
      out << a.data() << ' ';
    out << '\n';
  }
}

template <class T> Node<T> *Node<T>::freeList = NULL;
//Node<GraphArc> *Node<GraphArc>::freeList = NULL;
template <class T>const long Node<T>::newBlocksize = 64; 
//const long Node<GraphArc>::newBlocksize = 64;////////////

//template <class T> Node<T> *Node<T>::freeList = NULL;
//Node<long> *Node<long>::freeList = NULL;
//template <class T> const long Node<T>::newBlocksize = 64;
//const long Node<long>::newBlocksize = 64;

struct pGraphArc {
  GraphArc *p;
  GraphArc * operator ->() const { return p; }
  operator GraphArc *() const { return p; }
};

long operator < (const pGraphArc l, const pGraphArc r) {
  return l->weight > r->weight;
}

// pGraphArc must be used rather than GraphArc * because in order to overload operator < "`operator <(const GraphArc *, const GraphArc *)' must have an argument of class or enumerated type"

struct GraphHeap {
  static List<GraphHeap *> usedBlocks;
  static GraphHeap *freeList;
  static const long newBlocksize;
  GraphHeap *left, *right;	// for balanced heap  
  long nDescend;
  GraphArc *arc;		// data at each vertex, or cross edge 
  pGraphArc *arcHeap;		// binary heap of sidetracks originating from a state
  long arcHeapSize;
  void *operator new(size_t s)
  {
    size_t dummy = s;
    dummy = dummy;
    GraphHeap *ret, *max;
    if (freeList) {
      ret = freeList;
      freeList = freeList->left;
      return ret;
    }
    freeList = (GraphHeap *)::operator new(newBlocksize * sizeof(GraphHeap));
    usedBlocks.push(freeList);
    freeList->left = NULL;
    max = freeList + newBlocksize - 1;
    for ( ret = freeList++; freeList < max ; ret = freeList++ )
      freeList->left = ret;
    return freeList--;
  }
  void operator delete(void *p) 
  {
    GraphHeap *e = (GraphHeap *)p;
    e->left = freeList;
    freeList = e;
  }
  static void freeAll()
  {
    while ( usedBlocks.notEmpty() ) {
      ::operator delete((void *)usedBlocks.top());
      usedBlocks.pop();
    }
    freeList = NULL;
  }
};

//template <class T> Node<T> *Node<T>::freeList = NULL;
//template <class T> const long Node<T>::newBlocksize = 64;
//Node<GraphHeap *> *Node<GraphHeap *>::freeList = NULL;
//const long Node<GraphHeap *>::newBlocksize = 64;

//Node<GraphArc *> *Node<GraphArc *>::freeList = NULL;
//const long Node<GraphArc *>::newBlocksize = 64;

//Node<List<GraphArc *> > *Node<List<GraphArc *> >::freeList = NULL;
//const long Node<List<GraphArc *> >::newBlocksize = 64;


List<GraphHeap *> GraphHeap::usedBlocks;
GraphHeap * GraphHeap::freeList = NULL;
const long GraphHeap::newBlocksize = 64;

long operator < (const GraphHeap &l, const GraphHeap &r) {
  return l.arc->weight > r.arc->weight;
}

struct EdgePath {
  GraphHeap *node;
  long heapPos;			// -1 if arc is GraphHeap.arc
  EdgePath *last;
  float weight;
};

long operator < (const EdgePath &l, const EdgePath &r) {
  return l.weight > r.weight;
}

Graph sidetrackGraph(Graph sourceG, Graph subtractG, float *dist);
void buildSidetracksHeap(long state, long pred);
void printTree(GraphHeap *t, long n);
Graph sidetracks;
GraphHeap **pathGraph;
GraphState *shortPathTree;

void insertShortPath(long source, long dest, ListIter<GraphArc *> &path)
{
  GraphArc *taken;
  for ( long iState = source ; iState != dest; iState = taken->dest ) {
    taken = &shortPathTree[iState].arcs.top();
    path.insert((GraphArc *)taken->data);
  }
}

List<List<GraphArc *> > *bestPaths(Graph graph, long source, long dest, long k)
{
  long nStates = graph.nStates;
  assert(nStates > 0 && graph.states);
  assert(source >= 0 && source < nStates);
  assert(dest >= 0 && dest < nStates);
  
  List<List<GraphArc *> > *paths = new List<List<GraphArc *> >;
  ListIter<List<GraphArc *> > insertHere(*paths); // append rather than push so best path comes first in list

  float *dist = new float[nStates];
  Graph shortPathGraph = shortestPathTree(graph, dest, dist);
  shortPathTree = shortPathGraph.states;
  
  if ( shortPathTree[source].arcs.notEmpty() || source == dest ) {

    ListIter<GraphArc *> path(insertHere.insert(List<GraphArc *>()));
    insertShortPath(source, dest, path);
  
    if ( k > 1 ) {
      GraphHeap::freeAll();
      List<List<GraphArc *> > graphPaths;
      Graph revPathTree = reverseGraph(shortPathGraph);
      pathGraph = new GraphHeap *[nStates];
      sidetracks = sidetrackGraph(graph, shortPathGraph, dist);
      bool *visited = new bool[nStates];
      for ( long i = 0 ; i < nStates ; ++i ) visited[i] = 0;
      depthFirstSearch(revPathTree, dest, visited, buildSidetracksHeap);
      if ( pathGraph[source] ) {
	//    for ( long i = 0 ; i < nStates ; ++i )
	//      printTree(pathGraph[i], 0);
	EdgePath *pathQueue = new EdgePath[4 * (k+1)];	// out-degree is at most 4
	EdgePath *endQueue = pathQueue;
	EdgePath *retired = new EdgePath[k+1];
	EdgePath *endRetired = retired;
	EdgePath newPath;
	newPath.weight = pathGraph[source]->arc->weight;
	newPath.heapPos = -1;
	newPath.node = pathGraph[source];
	newPath.last = NULL;
	heapAdd(pathQueue, endQueue++, newPath);
	while ( heapSize(pathQueue, endQueue) && --k ) {
	  EdgePath *top = pathQueue;
	  GraphArc *cutArc;
	  List<GraphArc *> shortPath;
	  //      cout << top->weight;
	  if ( top->heapPos == -1 )
	    cutArc = top->node->arc;
	  else
	    cutArc = top->node->arcHeap[top->heapPos];
	  shortPath.push( cutArc);
	  //      cout << ' ' << *cutArc;
	  EdgePath *last;
	  while ( (last = top->last) ) {
	    if ( !((last->heapPos == -1 && (top->heapPos == 0 || top->node == last->node->left || top->node == last->node->right )) || (last->heapPos >= 0 && top->heapPos != -1 )) ) { // got to p on a cross edge
	      if ( last->heapPos == -1 ) 
		cutArc = last->node->arc;
	      else
		cutArc = last->node->arcHeap[last->heapPos];
	      shortPath.push(cutArc);
	      //	  cout << ' ' << *cutArc;
	    }
	    top = last;
	  }
	  //      cout << '\n';
	  ListIter<GraphArc *> fullPath(insertHere.insert(List<GraphArc *>()));
	  long sourceState = source;
	  for ( ListIter<GraphArc *> cut(shortPath) ; cut ; ++cut ) {
	    insertShortPath(sourceState, cut.data()->source, fullPath);
	    sourceState = cut.data()->dest;
	    fullPath.insert((GraphArc *)cut.data()->data);
	  }
	  insertShortPath(sourceState, dest, fullPath);
	  *endRetired = pathQueue[0];
	  newPath.last = endRetired++;
	  heapPop(pathQueue, endQueue--);
	  long lastHeapPos = newPath.last->heapPos;
	  GraphArc *spawnVertex;
	  GraphHeap *from = newPath.last->node;
	  float lastWeight = newPath.last->weight;
	  if ( lastHeapPos == -1 ) {
	    spawnVertex = from->arc;
	    newPath.heapPos = -1;
	    if ( from->left ) {
	      newPath.node = from->left;
	      newPath.weight = lastWeight + (newPath.node->arc->weight - spawnVertex->weight);
	      heapAdd(pathQueue, endQueue++, newPath);
	    }
	    if ( from->right ) {
	      newPath.node = from->right;
	      newPath.weight = lastWeight + (newPath.node->arc->weight - spawnVertex->weight);
	      heapAdd(pathQueue, endQueue++, newPath);
	    }
	    if ( from->arcHeapSize ) {
	      newPath.heapPos = 0;
	      newPath.node = from;
	      newPath.weight = lastWeight + (newPath.node->arcHeap[0]->weight - spawnVertex->weight);
	      heapAdd(pathQueue, endQueue++, newPath);
	    }
	  } else {
	    spawnVertex = from->arcHeap[lastHeapPos];
	    newPath.node = from;
	    long iChild = 2 * lastHeapPos + 1;
	    if ( from->arcHeapSize > iChild  ) {
	      newPath.heapPos = iChild;
	      newPath.weight = lastWeight + (newPath.node->arcHeap[iChild]->weight - spawnVertex->weight);
	      heapAdd(pathQueue, endQueue++, newPath);
	      if ( from->arcHeapSize > ++iChild ) {
		newPath.heapPos = iChild;
		newPath.weight = lastWeight + (newPath.node->arcHeap[iChild]->weight - spawnVertex->weight);
		heapAdd(pathQueue, endQueue++, newPath);
	      }
	    }
	  }
	  if ( pathGraph[spawnVertex->dest] ) {
	    newPath.heapPos = -1;
	    newPath.node = pathGraph[spawnVertex->dest];
	    newPath.heapPos = -1;
	    newPath.weight = lastWeight + newPath.node->arc->weight;
	    heapAdd(pathQueue, endQueue++, newPath);       
	  }
	}
	delete[] pathQueue;
	delete[] retired;
      }
      GraphHeap::freeAll();
      delete[] pathGraph;
      delete[] visited;
      delete[] revPathTree.states;
      delete[] sidetracks.states;
    }
  }

  delete[] graph.states;
  delete[] shortPathGraph.states;
  delete[] dist;

  return paths;
}

// a sidetrack from a given state in an arc originating from any state along the shortest path to the destination, that is not in the shortest path tree.  Paths are uniquely determined by a sequence of sidetracks from the destination of the previous sidetrack, or the source state if there was no previous sidetrack (see Eppstein)

Graph sidetrackGraph(Graph fullGraph, Graph shortGraph, float *dist)
{
  //  subtracts shortGraph from fullGraph, arcs' data member points to arc in fullGraph
  assert(fullGraph.nStates == shortGraph.nStates);
  long nStates = fullGraph.nStates;
  GraphState *sub = new GraphState[nStates];
  for ( long i = 0 ; i < nStates ; ++i )
    if ( dist[i] != HUGE_VAL ) 
      for ( ListIter<GraphArc> l(fullGraph.states[i].arcs) ; l ; ++l ) {
	assert(i == l.data().source);
	long isShort = 0;
	for ( ListIter<GraphArc> r(shortGraph.states[i].arcs) ; r ; ++r )
	  if ( r.data().data == &l.data() ) { // arcs in shortest path tree have data pointing to the arc corresponding to it in the original graph
	    isShort = 1;
	    break;
	  }
	if ( !isShort )
	  if ( dist[l.data().dest] != HUGE_VAL ) {
	    GraphArc w = l.data();
	    w.weight = w.weight - (dist[i] - dist[w.dest]);
	    w.data = &l.data();
	    sub[i].arcs.push(w);
	  }
      }
  Graph ret;
  ret.nStates = fullGraph.nStates;
  ret.states = sub;
  return ret;
}

// see Eppstein's paper for explanation of this shared heap

void buildSidetracksHeap(long state, long pred)
{
  GraphHeap *prev;
  
  if ( pred == -1 )
    prev = NULL;
  else
    prev = pathGraph[pred];
  
  ListIter<GraphArc> s(sidetracks.states[state].arcs);
  if ( s ) {
    long heapSize = 0;
    GraphArc *min;
    min = &s.data();
    while ( ++s ) {
      if ( s.data().weight < min->weight )
	min = &s.data();
      ++heapSize;
    }
    pathGraph[state] = new GraphHeap;
    pathGraph[state]->arc = min;
    pathGraph[state]->arcHeapSize = heapSize;
    if ( heapSize ) {
      pGraphArc *heapStart = pathGraph[state]->arcHeap = new pGraphArc[heapSize];
      pGraphArc *heapI = heapStart;
      for ( ListIter<GraphArc> gArc(sidetracks.states[state].arcs) ; gArc ; ++gArc )
	if ( &gArc.data() != min )
	  (heapI++)->p = &gArc.data();
      assert(heapI == heapStart + heapSize);
      heapBuild(heapStart, heapStart + heapSize);
    } else
      pathGraph[state]->arcHeap = NULL;
    pathGraph[state] = newTreeHeapAdd(prev, pathGraph[state]);
  } else
    pathGraph[state] = prev;
}

// debugging print routines

void printTree(GraphHeap *t, long n)
{
  long i;
  for ( i = 0 ; i < n ; ++i ) cout << ' ';
  if ( !t ) {
    cout << "-\n";
    return;
  }
  cout << *t->arc;
  cout << " [";
  pGraphArc *heap = t->arcHeap;
  for ( i = 0 ; i < t->arcHeapSize ; ++i ) {
    cout << *heap[i].p;
  }
  cout << "]\n";
  if ( !t->left && !t->right )
    return;
  printTree(t->left, n+1);
  printTree(t->right, n+1);
}

void shortPrintTree(GraphHeap *t)
{
  cout << *t->arc;
  if ( !t->left && !t->right )
    return;
  cout << " (";
  if ( t->left) {
    shortPrintTree(t->left);
    if ( t->right ) {
      cout << ' ';
      shortPrintTree(t->right);
    }
  } else
    if ( t->right )
      shortPrintTree(t->right);
  cout << ')';
}      


